;===============================================================================
; A graphical analogue clock for CP/M 3.
;=============================================================================== 

.include "CPM.inc"

DialOuterRadius = 479
DialInnerRadius = 440
SecondHandLength = 200
MinuteHandLength = 180
HourHandLength = 100
TickOuterRadius = 225
TickInnerRadius = 210
NumberRadius = 245

.org CPM.TPA

	; Fetch the address of the BIOS.
	ld hl,($0001)
	
	; Offset it by the address of the CONOUT (function 4) jump table entry.
	ld de,3*(4-1) ; CONOUT = function 4.
	add hl,de
	
	; Store it as the VDU call vector.
	ld (VDU.Call),hl
	
	; Select VDU MODE 1.
	ld a,22 \ call VDU
	ld a,1 \ call VDU
	
	; Select VDU 5 (draw text as graphics).
	ld a,5 \ call VDU
	
	; Six dummy bytes.
	ld hl,0
	call VDU.16
	call VDU.16
	call VDU.16
	
	; Set the graphics origin to the centre of the screen.
	ld a,29 \ call VDU
	ld hl,1280/2 \ call VDU.16 ; Screen is 1280 units wide.
	ld hl,960/2 \ call VDU.16 ; Screen is 960 units high.

	; Set the foreground colour to grey.
	ld a,18 \ call VDU ; GCOL
	ld hl,0+63*256 \ call VDU.16

	; Fill the outer dial circle.
	ld hl,0
	ld de,0
	ld bc,DialOuterRadius
	call Circle.Fill
	
	; Set the foreground colour to black.
	ld a,18 \ call VDU
	ld hl,0 \ call VDU.16
	
	; Clear the inner dial circle.
	ld bc,DialInnerRadius
	call Circle.Fill
	
	; Restore the default colours.
	ld a,20
	call VDU

	; Draw the outer dial circle.
	ld bc,DialOuterRadius
	call Circle.Thick
	
	; Draw the inner dial circle.
	ld bc,DialInnerRadius
	call Circle.Thick

	; Draw the ticks and numbers onto the dial.
	ld hl,0 ; Start at 0 (12 o'clock).
	ld b,12 ; 12 ticks.
	ld c,0 ; Start from 0 (12 o'clock).	

-	push bc
	
	; Outer radius of the tick.
	ld a,TickOuterRadius
	push hl
	call PolarToCartesian
	call Move
	pop hl
	
	; Inner radius of the tick.
	ld a,TickInnerRadius
	push hl
	call PolarToCartesian
	call Draw	
	pop hl
	
	; Position the cursor to draw the number.
	ld a,NumberRadius
	push hl
	call PolarToCartesian
	
	; Retrieve BC (contains the current hour number).
	pop af
	pop bc
	push bc
	push af
	
	; If hours are zero, replace with 12.
	ld a,c
	or a
	jr nz,+
	ld a,12
+

	ld c,0 ; C is 10s
	cp 10
	jr c,+
	sub 10
	inc c
+
	ld b,a

	; CB = Hours (C=10s, B=Units).
	push bc
	
	; Offset the cursor horizontally to align the text correctly.
	ld a,c
	ld bc,-16 ; Hour is one digit, offset 4*4 pixels.
	or a
	jr z,+
	ld bc,-32 ; Hour is one digit, offset 8*4 pixels.
+	add hl,bc
	
	; Offset the cursor vertically to align the text correctly.
	ex de,hl
	ld bc,16 ; Hour is always one line, so offset 4*4 pixels.
	add hl,bc
	ex de,hl
	
	; Move the cursor into position.
	call Move
	
	; Draw the number.
	pop bc
	
	; Tens (if applicable).
	ld a,c
	or a
	jr z,+
	add a,'0'
	call VDU
+
	; Units.
	ld a,b
	add a,'0'
	call VDU
	
	pop hl
	
	; Advance 30 degrees.
	ld de,30
	add hl,de
	
	pop bc
	; Advance to the next hour.
	inc c
	djnz -
	
	; Draw the dots around the inside of the dial.
	ld hl,0 ; Start at 0 (12 o'clock).
	ld b,60 ; 60 dots.

-	push bc
	
	; Outer radius of the tick.
	ld a,TickInnerRadius
	push hl
	call PolarToCartesian
	
	ld a,25 \ call VDU ; PLOT
	ld a,69 \ call VDU ; Plot single point.
	call VDU.16 ; X
	ex de,hl
	call VDU.16 ; Y	
	
	; Advance 6 degrees.
	pop hl
	ld de,6
	add hl,de
	
	pop bc
	djnz -
	
ClockLoop
	
	; Fetch the current time.
	call FetchTime

	; Calculate the angles of the hands.
	ld hl,(CurrentSecondHandAngle)
	ld (PreviousSecondHandAngle),hl
	ld hl,(Time.Seconds)
	ld h,0
	call Multiply.6
	ld (CurrentSecondHandAngle),hl
	
	ld hl,(CurrentMinuteHandAngle)
	ld (PreviousMinuteHandAngle),hl
	ld hl,(Time.Minutes)
	ld h,0
	call Multiply.6
	; Offset by seconds / 10 to position the hand more accurately.
	ld a,(Time.Seconds.Packed)
	srl a \ srl a \ srl a \ srl a
	ld e,a
	ld d,0
	add hl,de
	ld (CurrentMinuteHandAngle),hl
	
	ld hl,(CurrentHourHandAngle)
	ld (PreviousHourHandAngle),hl
	ld hl,(Time.Hours)
	ld h,0
	call Multiply.30
	; Offset by minutes / 2 to position the hand more accurately.
	ld a,(Time.Minutes)
	sra a
	ld e,a
	ld d,0
	add hl,de
	ld (CurrentHourHandAngle),hl	

	; Erase the old hands.
	call Home
	ld hl,(PreviousSecondHandAngle)
	ld de,(CurrentSecondHandAngle)
	call CompareHLDE
	jr z,+
	ld a,SecondHandLength
	call PolarToCartesian
	call Erase
+
	
	call Home
	ld hl,(PreviousMinuteHandAngle)
	ld de,(CurrentMinuteHandAngle)
	call CompareHLDE
	jr z,+
	ld a,MinuteHandLength
	call PolarToCartesian
	call Erase
+

	call Home
	ld hl,(PreviousHourHandAngle)
	ld de,(CurrentHourHandAngle)
	call CompareHLDE
	jr z,+
	ld a,HourHandLength
	call PolarToCartesian
	call Erase
+
	; Draw the new hands.
	call Home
	ld hl,(CurrentSecondHandAngle)
	ld a,SecondHandLength
	call PolarToCartesian
	call Draw
	
	call Home
	ld hl,(CurrentMinuteHandAngle)
	ld a,MinuteHandLength
	call PolarToCartesian
	call Draw
	
	call Home
	ld hl,(CurrentHourHandAngle)
	ld a,HourHandLength
	call PolarToCartesian
	call Draw
	
	; Are we finished?
	ld c,CPM.BDOS.C_RAWIO
	ld e,$FF
	call CPM.BDOS
	cp 27
	jp nz,ClockLoop
	
	; Reselect VDU mode 0.
	ld a,22 \ call VDU
	xor a \ call VDU
	
	rst $00

; Angles of the hands in the current frame.
CurrentSecondHandAngle
	.dw 0
CurrentMinuteHandAngle
	.dw 0
CurrentHourHandAngle
	.dw 0

; Angles of the hands in the previous frame.
PreviousSecondHandAngle
	.dw 0
PreviousMinuteHandAngle
	.dw 0
PreviousHourHandAngle
	.dw 0

;------------------------------------------------------------------------------- 
; Outputs the value in A to the VDU directly.
;------------------------------------------------------------------------------- 
VDU
	push bc
	push de
	push hl
	ld c,a
VDU.Call = $+1
	call $0000
	pop hl
	pop de
	pop bc
	ret

;------------------------------------------------------------------------------- 
; Outputs the value in HL to the VDU directly.
;------------------------------------------------------------------------------- 
VDU.16
	ld a,l
	call VDU
	ld a,h
	jp VDU

;------------------------------------------------------------------------------- 
; Moves the graphics cursor to (0,0).
;------------------------------------------------------------------------------- 
Home
	ld hl,0
	ld de,0
	; Deliberate run-on.

;------------------------------------------------------------------------------- 
; Moves the graphics cursor to (HL,DE).
;------------------------------------------------------------------------------- 
Move
	; Move to the origin of the circle.
	ld a,25 \ call VDU ; PLOT
	ld a,4 \ call VDU ; MOVE
	; X
	call VDU.16
	; Y
	ex de,hl
	call VDU.16
	ex de,hl
	ret
	
;------------------------------------------------------------------------------- 
; Draws a line to (HL,DE).
;------------------------------------------------------------------------------- 
Draw
	ld a,25 \ call VDU ; PLOT
	ld a,5 \ call VDU ; DRAW
	; X
	call VDU.16
	; Y
	ex de,hl
	call VDU.16
	ex de,hl
	ret

;------------------------------------------------------------------------------- 
; Erases a line to (HL,DE).
;------------------------------------------------------------------------------- 
Erase
	ld a,25 \ call VDU ; PLOT
	ld a,7 \ call VDU ; Draw line in background colour.
	; X
	call VDU.16
	; Y
	ex de,hl
	call VDU.16
	ex de,hl
	ret

;------------------------------------------------------------------------------- 
; Draws a circle at (HL,DE) with a radius of BC.
;------------------------------------------------------------------------------- 
Circle
	call Move
	
	; Draw the circle.
	ld a,25 \ call VDU ; PLOT
	ld a,145 \ call VDU ; Circle
	
	; Radius
	push hl
	ld h,b
	ld l,c
	call VDU.16
	ld hl,0
	call VDU.16
	pop hl
	ret

;------------------------------------------------------------------------------- 
; Draws a thickly outlined circle at (HL,DE) with a radius of BC.
;------------------------------------------------------------------------------- 
Circle.Thick
	; Draw the circle at the desired position.
	call Circle
	; Draw the circle one to the left.
	dec hl
	call Circle
	; Draw the circle one to the left and one down.
	dec de
	call Circle
	; Draw the circle one down.
	inc hl
	call Circle
	; Restore the original coordinates.
	inc de
	ret

;------------------------------------------------------------------------------- 
; Fills a circle at (HL,DE) with a radius of BC.
;------------------------------------------------------------------------------- 
Circle.Fill
	call Move
	
	; Draw the circle.
	ld a,25 \ call VDU ; PLOT
	ld a,153 \ call VDU ; Circle fill
	
	; Radius
	push hl
	ld h,b
	ld l,c
	call VDU.16
	ld hl,0
	call VDU.16
	pop hl
	ret

;------------------------------------------------------------------------------- 
; Generates a trig table used to calculate the position of clock hands.
;------------------------------------------------------------------------------- 
AngleTable
	.for a = 0 to 359
		.dw 420 * sin(a / 180 * pi())
	.loop

;------------------------------------------------------------------------------- 
; Computes AHL = DE*A (unsigned).
;------------------------------------------------------------------------------- 
Multiply.U16U8
	ld hl,0
	ld c,0
	add a,a
	jr nc,$+4
	ld h,d
	ld l,e

	.rept 7
		add hl,hl
		rla
		jr nc,$+4
		add hl,de
		adc a,c
	.loop
	ret

;------------------------------------------------------------------------------- 
; Scales DE (signed) by A (0 to 255).
;------------------------------------------------------------------------------- 
ScaleDE
	bit 7,d
	jr nz,+
	call Multiply.U16U8
	ld d,a
	ld e,h
	ret
+	ld h,a
	ld a,d \ cpl \ ld d,a
	ld a,e \ cpl \ ld e,a
	inc de
	ld a,h
	call Multiply.U16U8
	cpl \ ld d,a
	ld a,h \ cpl \ ld e,a
	inc de
	ret

;------------------------------------------------------------------------------- 
; Computes HL = HL % 360 (HL must be > 0).
;------------------------------------------------------------------------------- 
Mod360
	push de
-	push hl
	ld de,-360
	add hl,de
	jr nc,Mod360.Done
	pop de
	jr -
Mod360.Done
	pop hl
	pop de
	ret	

;------------------------------------------------------------------------------- 
; Multiplies HL by 6.
;------------------------------------------------------------------------------- 
Multiply.6
	; *6 = *2 + *4
	push de
	add hl,hl
	push hl ; *2
	add hl,hl
	pop de ; *2
	add hl,de
	pop de
	ret
	
;------------------------------------------------------------------------------- 
; Multiplies HL by 30.
;------------------------------------------------------------------------------- 
Multiply.30
	; *30 = *2 + *4 + *8 + *16
	push de
	add hl,hl
	push hl ; *2
	add hl,hl
	push hl ; *4
	add hl,hl
	push hl ; *8
	add hl,hl
	pop de ; *8
	add hl,de
	pop de ; *4
	add hl,de
	pop de ; *2
	add hl,de
	pop de
	ret

;------------------------------------------------------------------------------- 
; Converts polar coordinates (HL=angle, A=radius) to cartesian ones (HL, DE).
;------------------------------------------------------------------------------- 
PolarToCartesian
	; Store the angle and radius.
	push hl
	push af
	
	; X = r*sin(t)
	
	; Fetch sin(t) from the angle table.
	call Mod360
	add hl,hl
	ld de,AngleTable
	add hl,de
	ld e,(hl)
	inc hl
	ld d,(hl)
	; DE = sin(t)
	
	; Scale the result.
	call ScaleDE
	
	; Store the result.
	ld (PolarToCartesian.Result.X),de
	
	; Restore the radius and angle.
	pop af
	pop hl
	
	; Y = r*cos(t)
	
	; cos(t) = sin(t+90)
	ld de,90
	add hl,de
	call Mod360
	add hl,hl
	ld de,AngleTable
	add hl,de
	ld e,(hl)
	inc hl
	ld d,(hl)
	; DE = cos(t)
	
	; Scale the result.
	call ScaleDE
	
PolarToCartesian.Result.X = $+1
	ld hl,$0000
	ret
	
;------------------------------------------------------------------------------- 
; Reads the value from the SCB offset by A into A/HL.
;------------------------------------------------------------------------------- 
ReadSCB
	ld (ReadSCB.Offset),a
	ld c,CPM.BDOS.S_SYSVAR
	ld de,ReadSCB.Offset
	jp CPM.BDOS
ReadSCB.Offset
	.db 0
ReadSCB.Command
	.db 0

;------------------------------------------------------------------------------- 
; Unpack a binary coded decimal in A to binary.
;------------------------------------------------------------------------------- 
UnpackBCD
	push bc
	push af
	srl a
	srl a
	srl a
	srl a
	ld b,a
	or a
	jr z,+
	xor a
-	add a,10
	djnz -
+	ld b,a
	pop af
	and $0F
	add a,b
	pop bc
	ret

;------------------------------------------------------------------------------- 
; Fetches the current time.
;------------------------------------------------------------------------------- 
FetchTime
	; Fetch the address of the BIOS.
	ld hl,($0001)
	
	; Offset it by the address of the TIME (function 26) jump table entry.
	ld de,3*(26-1) ; TIME = function 6.
	add hl,de
	
	; Invoke the TIME BIOS function.
	ld c,0
	call CallHL
	
	; In case the time is updated by an ISR, temporary disable interrupts.
	di
	
	; Fetch the time.
	ld a,CPM.SCB.SEC
	call ReadSCB
	ld (Time.Seconds.Packed),a
	call UnpackBCD
	ld (Time.Seconds),a
	
	ld a,CPM.SCB.MIN
	call ReadSCB
	ld (Time.Minutes.Packed),a
	call UnpackBCD
	ld (Time.Minutes),a
	
	ld a,CPM.SCB.HOUR
	call ReadSCB
	ld (Time.Hours.Packed),a
	call UnpackBCD
	ld (Time.Hours),a
	
	; Re-enable interrupts.
	ei	
	ret

Time.Seconds.Packed
	.db 0
Time.Minutes.Packed
	.db 0
Time.Hours.Packed
	.db 0

Time.Seconds
	.db 0
Time.Minutes
	.db 0
Time.Hours
	.db 0

;------------------------------------------------------------------------------- 
; Call this to call the address in HL.
;------------------------------------------------------------------------------- 
CallHL
	jp (hl)

;------------------------------------------------------------------------------- 
; Compares HL and DE.
;------------------------------------------------------------------------------- 
CompareHLDE
	push hl
	push bc
	ld b,a
	xor a
	sbc hl,de
	ld a,b
	pop bc
	pop hl
	ret